#ifndef EL__OSDEP_SYSNAME_H
#define EL__OSDEP_SYSNAME_H

#ifdef __cplusplus
extern "C" {
#endif

extern unsigned char system_name[];

extern void get_system_name(void);

#ifdef __cplusplus
}
#endif

#endif
