
#ifndef EL__ECMASCRIPT_SPIDERMONKEY_UNIBAR_H
#define EL__ECMASCRIPT_SPIDERMONKEY_UNIBAR_H

#include "ecmascript/spidermonkey/util.h"

extern JSClass menubar_class;
extern JSClass statusbar_class;
extern JSPropertySpec unibar_props[];

#endif
